// kummer.h

#ifndef KUMMER_H
#define KUMMER_H

#include <ginac/ginac.h>

class kummer_point;
class kummer_point_gcd;

// Kummer surface
class kummer
{
private:
	GiNaC::numeric f_coeff[7];
	GiNaC::ex def_eq;
	GiNaC::ex biquadratic_form[4][4];
	GiNaC::ex duplication[4];
public:
	kummer(const GiNaC::numeric * const coeff);
	const GiNaC::numeric& get_coeff(const int i) const
		{ return f_coeff[i]; }
	const GiNaC::ex& get_def_eq() const
		{ return def_eq; }
	const GiNaC::ex& get_biquadratic_form(const int i, const int j) const
		{ return biquadratic_form[i-1][j-1]; }
	const GiNaC::ex& get_duplication(const int i) const
		{ return duplication[i-1]; }
	const kummer_point duplicate(const kummer_point& p) const;
	const kummer_point duplicate_numerically(const kummer_point& p) const;
	const kummer_point pseudo_add(const kummer_point& a, const kummer_point& b, const kummer_point& p) const;
};

// point on a Kummer surface
class kummer_point
{
	friend class kummer_point_gcd;
private:
	GiNaC::numeric coord[4];
public:
	kummer_point() { coord[0] = coord[1] = coord[2] = 0; coord[3] = 1; }
	kummer_point(const kummer_point& p);
	kummer_point(const GiNaC::numeric * const c);
	const GiNaC::numeric& get_coord(const int i) const
		{ return coord[i-1]; }
	const kummer_point normalize() const;
	void normalize_numerically();
	bool is_zero_mod_p(const GiNaC::numeric & prime) const
		{ return mod(coord[0], prime).is_zero()
			&& mod(coord[1], prime).is_zero()
			&& mod(coord[2], prime).is_zero(); }
	const GiNaC::numeric naive_height() const;
	const GiNaC::numeric canonical_height(const kummer& k) const;
	const GiNaC::numeric canonical_height_Stoll(const kummer& k) const;
private:
	int ord_p(const GiNaC::numeric& prime) const;
};

// point on a Kummer surface with the greatest common divisor of the coordinates
class kummer_point_gcd {
public:
	kummer_point p;
	GiNaC::numeric d;
public:
	kummer_point_gcd() { d = 1; }
	kummer_point_gcd(const kummer_point& kp);
	kummer_point_gcd(const kummer_point_gcd& kpg) { p = kpg.p; d = kpg.d; }
	const kummer_point get_actual_point() const;
};

bool is_on_kummer(const kummer& k, const kummer_point& p);

GiNaC::lst ifactor(const GiNaC::ex &e);

#endif
